"""Test the conversion of Markdown pages.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os
from unittest.mock import patch

import pytest

from incorporealcms.markdown import (generate_parent_navs, handle_markdown_file_path,
                                     instance_resource_path_to_request_path, parse_md,
                                     request_path_to_breadcrumb_display)

HERE = os.path.dirname(os.path.abspath(__file__))
os.chdir(os.path.join(HERE, 'instance/', 'pages/'))


def test_generate_page_navs_index():
    """Test that the index page has navs to the root (itself)."""
    assert generate_parent_navs('index.md') == [('example.org', '/')]


def test_generate_page_navs_subdir_index():
    """Test that dir pages have navs to the root and themselves."""
    assert generate_parent_navs('subdir/index.md') == [('example.org', '/'), ('subdir', '/subdir/')]


def test_generate_page_navs_subdir_real_page():
    """Test that real pages have navs to the root, their parent, and themselves."""
    assert generate_parent_navs('subdir/page.md') == [('example.org', '/'), ('subdir', '/subdir/'),
                                                      ('Page', '/subdir/page')]


def test_generate_page_navs_subdir_with_title_parsing_real_page():
    """Test that title metadata is used in the nav text."""
    assert generate_parent_navs('subdir-with-title/page.md') == [
        ('example.org', '/'),
        ('SUB!', '/subdir-with-title/'),
        ('page', '/subdir-with-title/page')
    ]


def test_generate_page_navs_subdir_with_no_index():
    """Test that breadcrumbs still generate even if a subdir doesn't have an index.md."""
    assert generate_parent_navs('no-index-dir/page.md') == [
        ('example.org', '/'),
        ('/no-index-dir/', '/no-index-dir/'),
        ('page', '/no-index-dir/page')
    ]


def test_page_includes_themes_with_default():
    """Test that a request contains the configured themes and sets the default as appropriate."""
    assert '<link rel="stylesheet" type="text/css" title="light" href="/static/css/light.css">'\
        in handle_markdown_file_path('index.md')
    assert '<link rel="alternate stylesheet" type="text/css" title="dark" href="/static/css/dark.css">'\
        in handle_markdown_file_path('index.md')
    assert '<a href="" onclick="setStyle(\'light\'); return false;">[light]</a>'\
        in handle_markdown_file_path('index.md')
    assert '<a href="" onclick="setStyle(\'dark\'); return false;">[dark]</a>'\
        in handle_markdown_file_path('index.md')


def test_render_with_style_overrides():
    """Test that the default can be changed."""
    with patch('incorporealcms.Config.DEFAULT_PAGE_STYLE', 'dark'):
        assert '<link rel="stylesheet" type="text/css" title="dark" href="/static/css/dark.css">'\
            in handle_markdown_file_path('index.md')
        assert '<link rel="alternate stylesheet" type="text/css" title="light" href="/static/css/light.css">'\
            in handle_markdown_file_path('index.md')
    assert '<a href="" onclick="setStyle(\'light\'); return false;">[light]</a>'\
        in handle_markdown_file_path('index.md')
    assert '<a href="" onclick="setStyle(\'dark\'); return false;">[dark]</a>'\
        in handle_markdown_file_path('index.md')


def test_render_with_default_style_override():
    """Test that theme overrides work, and if a requested theme doesn't exist, the default is loaded."""
    with patch('incorporealcms.Config.PAGE_STYLES', {'cool': '/static/css/cool.css',
                                                     'warm': '/static/css/warm.css'}):
        with patch('incorporealcms.Config.DEFAULT_PAGE_STYLE', 'warm'):
            assert '<link rel="stylesheet" type="text/css" title="warm" href="/static/css/warm.css">'\
                in handle_markdown_file_path('index.md')
            assert '<link rel="alternate stylesheet" type="text/css" title="cool" href="/static/css/cool.css">'\
                in handle_markdown_file_path('index.md')
            assert '<link rel="alternate stylesheet" type="text/css" title="light" href="/static/css/light.css">'\
                not in handle_markdown_file_path('index.md')
            assert '<a href="" onclick="setStyle(\'warm\'); return false;">[warm]</a>'\
                in handle_markdown_file_path('index.md')
            assert '<a href="" onclick="setStyle(\'cool\'); return false;">[cool]</a>'\
                in handle_markdown_file_path('index.md')


def test_redirects_error_unsupported():
    """Test that we throw a warning about the barely-used Markdown redirect tag, which we can't support via SSG."""
    os.chdir(os.path.join(HERE, 'instance/', 'broken/'))
    with pytest.raises(NotImplementedError):
        handle_markdown_file_path('redirect.md')
    os.chdir(os.path.join(HERE, 'instance/', 'pages/'))


def test_instance_resource_path_to_request_path_on_index():
    """Test index.md -> /."""
    assert instance_resource_path_to_request_path('index.md') == '/'


def test_instance_resource_path_to_request_path_on_page():
    """Test no-title.md -> no-title."""
    assert instance_resource_path_to_request_path('no-title.md') == '/no-title'


def test_instance_resource_path_to_request_path_on_subdir():
    """Test subdir/index.md -> subdir/."""
    assert instance_resource_path_to_request_path('subdir/index.md') == '/subdir/'


def test_instance_resource_path_to_request_path_on_subdir_and_page():
    """Test subdir/page.md -> subdir/page."""
    assert instance_resource_path_to_request_path('subdir/page.md') == '/subdir/page'


def test_request_path_to_breadcrumb_display_patterns():
    """Test various conversions from request path to leaf nodes for display in the breadcrumbs."""
    assert request_path_to_breadcrumb_display('/foo') == 'foo'
    assert request_path_to_breadcrumb_display('/foo/') == 'foo'
    assert request_path_to_breadcrumb_display('/foo/bar') == 'bar'
    assert request_path_to_breadcrumb_display('/foo/bar/') == 'bar'
    assert request_path_to_breadcrumb_display('/') == ''


def test_parse_md_metadata():
    """Test the direct results of parsing a markdown file."""
    content, md, page_name, page_title, mtime = parse_md('more-metadata.md')
    assert page_name == 'title for the page'
    assert page_title == 'title for the page - example.org'


def test_parse_md_metadata_forced_no_title():
    """Test the direct results of parsing a markdown file."""
    content, md, page_name, page_title, mtime = parse_md('forced-no-title.md')
    assert page_name == ''
    assert page_title == 'example.org'


def test_parse_md_metadata_no_title_so_path():
    """Test the direct results of parsing a markdown file."""
    content, md, page_name, page_title, mtime = parse_md('subdir/index.md')
    assert page_name == '/subdir/'
    assert page_title == '/subdir/ - example.org'


def test_parse_md_no_file():
    """Test the direct results of parsing a markdown file."""
    with pytest.raises(FileNotFoundError):
        content, md, page_name, page_title, mtime = parse_md('nope.md')


def test_parse_md_bad_file():
    """Test the direct results of parsing a markdown file."""
    with pytest.raises(ValueError):
        content, md, page_name, page_title, mtime = parse_md('actually-a-png.md')



"""Test command line invocations.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os
import tempfile
from subprocess import run

HERE = os.path.dirname(os.path.abspath(__file__))


def test_build():
    """Test some of the output of the core builder command."""
    with tempfile.TemporaryDirectory() as tmpdir:
        result = run(['incorporealcms-build', os.path.join(HERE, 'instance'), tmpdir],
                     capture_output=True, encoding='utf8')
        assert "creating temporary directory" in result.stdout
        assert "copying file" in result.stdout
        assert "creating symlink" in result.stdout
        assert "creating directory" in result.stdout
        assert "renaming" in result.stdout


def test_build_error():
    """Test some of the output of the core builder command."""
    result = run(['incorporealcms-build', os.path.join(HERE, 'instance'), os.path.join(HERE, 'test_markdown.py')],
                 capture_output=True, encoding='utf8')
    assert "specified output path" in result.stderr
    assert "exists as a file!" in result.stderr



"""An application for running my Markdown-based sites.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import json
import logging
import os
from logging.config import dictConfig

from jinja2 import Environment, PackageLoader, select_autoescape
from termcolor import cprint

from incorporealcms.config import Config

jinja_env = Environment(
    loader=PackageLoader('incorporealcms'),
    autoescape=select_autoescape(),
)

# dynamically generate version number
try:
    # packaged/pip install -e . value
    from ._version import version as __version__
except ImportError:
    # local clone value
    from setuptools_scm import get_version
    __version__ = get_version(root='..', relative_to=__file__)


def init_instance(instance_path: str, extra_config: dict = None):
    """Create the instance context, with allowances for customizing path and test settings."""
    # load the instance config.json, if there is one
    instance_config = os.path.join(instance_path, 'config.json')
    try:
        with open(instance_config, 'r') as config:
            config_dict = json.load(config)
            cprint(f"splicing {config_dict} into the config", 'yellow')
            Config.update(config_dict)
    except OSError:
        raise ValueError("instance path does not seem to be a site instance!")

    if extra_config:
        cprint(f"splicing {extra_config} into the config", 'yellow')
        Config.update(extra_config)

    # stash some stuff
    Config.INSTANCE_DIR = os.path.abspath(instance_path)

    dictConfig(Config.LOGGING)
    logger = logging.getLogger(__name__)

    logger.debug("instance dir: %s", Config.INSTANCE_DIR)



"""Generate Atom and RSS feeds based on content in a blog-ish location.

This parses a special root directory, feed/, for YYYYMMDD-foo.md files,
and combines them into an Atom or RSS feed. These files *should* be symlinks
to the real pages, which may mirror the same YYYYMMDD-foo.md file naming scheme
under pages/ (which may make sense for a blog) if they want, but could just
as well be pages/foo content.

SPDX-FileCopyrightText: © 2023 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import logging
import os
import re

from feedgen.feed import FeedGenerator

from incorporealcms.config import Config
from incorporealcms.markdown import instance_resource_path_to_request_path, parse_md

logger = logging.getLogger(__name__)


def generate_feed(feed_type: str, instance_dir: str, dest_dir: str) -> None:
    """Generate the Atom or RSS feed as requested.

    Args:
        feed_type: 'atom' or 'rss' feed
        instance_dir: the directory for the instance, containing both the feed dir and pages
        dest_dir: the directory to place the feed subdir and requested feed
    """
    fg = FeedGenerator()
    fg.id(f'https://{Config.DOMAIN_NAME}/')
    fg.title(f'{Config.TITLE_SUFFIX}')
    fg.author(Config.AUTHOR)
    fg.link(href=f'https://{Config.DOMAIN_NAME}/feed/{feed_type}', rel='self')
    fg.link(href=f'https://{Config.DOMAIN_NAME}', rel='alternate')
    fg.subtitle(f"Blog posts and other interesting materials from {Config.TITLE_SUFFIX}")

    # get recent feeds
    feed_path = os.path.join(instance_dir, 'feed')
    feed_entry_paths = [os.path.join(dirpath, filename) for dirpath, _, filenames in os.walk(feed_path)
                        for filename in filenames if os.path.islink(os.path.join(dirpath, filename))]
    for feed_entry_path in sorted(feed_entry_paths):
        # get the actual file to parse it
        os.chdir(os.path.abspath(os.path.join(instance_dir, 'pages')))
        resolved_path = os.path.relpath(os.path.realpath(feed_entry_path), os.path.join(instance_dir, 'pages'))
        try:
            content, md, page_name, page_title, mtime = parse_md(resolved_path)
            link = f'https://{Config.DOMAIN_NAME}{instance_resource_path_to_request_path(resolved_path)}'
        except (OSError, ValueError, TypeError):
            logger.exception("error loading/rendering markdown!")
            raise

        fe = fg.add_entry()
        fe.id(_generate_feed_id(feed_entry_path, instance_resource_path_to_request_path(resolved_path)))
        fe.title(page_title)
        fe.author(Config.AUTHOR)
        fe.link(href=link)
        fe.content(content, type='html')

    if feed_type == 'rss':
        try:
            os.mkdir(os.path.join(dest_dir, 'feed'))
        except FileExistsError:
            pass
        with open(os.path.join(dest_dir, 'feed', 'rss'), 'wb') as feed_file:
            feed_file.write(fg.rss_str(pretty=True))
    else:
        try:
            os.mkdir(os.path.join(dest_dir, 'feed'))
        except FileExistsError:
            pass
        with open(os.path.join(dest_dir, 'feed', 'atom'), 'wb') as feed_file:
            feed_file.write(fg.atom_str(pretty=True))


def _generate_feed_id(feed_entry_path, request_path):
    """For a relative file path, generate the Atom/RSS feed ID for it."""
    date = re.sub(r'.*(\d{4})(\d{2})(\d{2}).*', r'\1-\2-\3', feed_entry_path)
    cleaned = request_path.replace('#', '/')
    return f'tag:{Config.DOMAIN_NAME},{date}:{cleaned}'



"""Test basic configuration stuff.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os

import pytest

from incorporealcms import init_instance
from incorporealcms.config import Config

HERE = os.path.dirname(os.path.abspath(__file__))


def test_config():
    """Test that the app initialization sets values not normally present in the config."""
    # this may have gotten here from other imports in other tests
    try:
        delattr(Config, 'INSTANCE_VALUE')
    except AttributeError:
        pass

    assert not getattr(Config, 'INSTANCE_VALUE', None)
    assert not getattr(Config, 'EXTRA_VALUE', None)

    instance_path = os.path.join(HERE, 'instance')
    init_instance(instance_path=instance_path, extra_config={"EXTRA_VALUE": "hello"})

    assert getattr(Config, 'INSTANCE_VALUE', None) == "hi"
    assert getattr(Config, 'EXTRA_VALUE', None) == "hello"


def test_broken_config():
    """Test that the app initialization errors when not given an instance-looking thing."""
    with pytest.raises(ValueError):
        instance_path = os.path.join(HERE, 'blah')
        init_instance(instance_path=instance_path)



"""Test the high level SSG operations.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os
import tempfile

import incorporealcms.ssg as ssg
from incorporealcms import init_instance

HERE = os.path.dirname(os.path.abspath(__file__))

instance_dir = os.path.join(HERE, 'instance')
init_instance(instance_dir)


def test_file_copy():
    """Test the ability to sync a file to the output dir."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(instance_dir, 'pages'))
        generator.build_file_in_destination(os.path.join(instance_dir, 'pages'), '', 'no-title.md', tmpdir,
                                            True)
        assert os.path.exists(os.path.join(tmpdir, 'no-title.md'))
        assert os.path.exists(os.path.join(tmpdir, 'no-title.html'))


def test_file_copy_no_markdown():
    """Test the ability to sync a file to the output dir."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(instance_dir, 'pages'))
        generator.build_file_in_destination(os.path.join(instance_dir, 'pages'), '', 'no-title.md', tmpdir,
                                            False)
        assert os.path.exists(os.path.join(tmpdir, 'no-title.md'))
        assert not os.path.exists(os.path.join(tmpdir, 'no-title.html'))


def test_file_copy_symlink():
    """Test the ability to sync a file to the output dir."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(instance_dir, 'pages'))
        generator.build_file_in_destination(os.path.join(instance_dir, 'pages'), '', 'symlink-to-foo.txt', tmpdir)
        # need to copy the destination for os.path.exists to be happy with this
        generator.build_file_in_destination(os.path.join(instance_dir, 'pages'), '', 'foo.txt', tmpdir)
        assert os.path.exists(os.path.join(tmpdir, 'symlink-to-foo.txt'))
        assert os.path.islink(os.path.join(tmpdir, 'symlink-to-foo.txt'))


def test_dir_copy():
    """Test the ability to sync a directory to the output dir."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(instance_dir, 'pages'))
        generator.build_subdir_in_destination(os.path.join(instance_dir, 'pages'), '', 'media', tmpdir)
        assert os.path.exists(os.path.join(tmpdir, 'media'))
        assert os.path.isdir(os.path.join(tmpdir, 'media'))


def test_dir_copy_symlink():
    """Test the ability to sync a directory to the output dir."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(instance_dir, 'pages'))
        generator.build_subdir_in_destination(os.path.join(instance_dir, 'pages'), '', 'symlink-to-subdir', tmpdir)
        # need to copy the destination for os.path.exists to be happy with this
        generator.build_subdir_in_destination(os.path.join(instance_dir, 'pages'), '', 'subdir', tmpdir)
        assert os.path.exists(os.path.join(tmpdir, 'symlink-to-subdir'))
        assert os.path.isdir(os.path.join(tmpdir, 'symlink-to-subdir'))
        assert os.path.islink(os.path.join(tmpdir, 'symlink-to-subdir'))


def test_build_in_destination():
    """Test the ability to walk a source and populate the destination."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generator = ssg.StaticSiteGenerator(src_dir, tmpdir)
        generator.build_in_destination(os.path.join(src_dir, 'pages'), tmpdir)

        assert os.path.exists(os.path.join(tmpdir, 'index.md'))
        assert os.path.exists(os.path.join(tmpdir, 'index.html'))
        assert os.path.exists(os.path.join(tmpdir, 'subdir', 'index.md'))
        assert os.path.exists(os.path.join(tmpdir, 'subdir', 'index.html'))



"""Process Markdown pages.

With the project now being a SSG, most files we just let the web server serve
as is, but .md files need to be processed with a Markdown parser, so a lot of this
is our tweaks and customizations for pages my way.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import datetime
import logging
import os
import re

import markdown
from markupsafe import Markup

from incorporealcms import jinja_env
from incorporealcms.config import Config

logger = logging.getLogger(__name__)


def get_meta_str(md, key):
    """Provide the page's (parsed in Markup obj md) metadata for the specified key, or '' if unset."""
    return " ".join(md.Meta.get(key)) if md.Meta.get(key) else ""


def init_md():
    """Initialize the Markdown parser.

    This used to done at the app level in __init__, but extensions like footnotes apparently
    assume the parser to only live for the length of parsing one document, and create double
    footnote ref links if the one parser sees the same document multiple times.
    """
    # initialize markdown parser from config, but include
    # extensions our app depends on, like the meta extension
    return markdown.Markdown(extensions=Config.MARKDOWN_EXTENSIONS + ['meta'],
                             extension_configs=Config.MARKDOWN_EXTENSION_CONFIGS)


def instance_resource_path_to_request_path(path):
    """Reverse a relative disk path to the path that would show up in a URL request."""
    return '/' + re.sub(r'.md$', '', re.sub(r'index.md$', '', path))


def parse_md(path: str):
    """Given a file to parse, return file content and other derived data along with the md object.

    Args:
        path: the path to the file to render
    """
    try:
        logger.debug("opening path '%s'", path)
        with open(path, 'r') as input_file:
            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(input_file.name), tz=datetime.timezone.utc)
            entry = input_file.read()
        logger.debug("path '%s' read", path)
        md = init_md()
        content = Markup(md.convert(entry))     # nosec B704
    except (OSError, FileNotFoundError):
        logger.exception("path '%s' could not be opened!", path)
        raise
    except ValueError:
        logger.exception("error parsing/rendering markdown!")
        raise

    logger.debug("file metadata: %s", md.Meta)

    page_name = get_meta_str(md, 'title') if md.Meta.get('title') else instance_resource_path_to_request_path(path)
    page_title = f'{page_name} - {Config.TITLE_SUFFIX}' if page_name else Config.TITLE_SUFFIX
    logger.debug("title (potentially derived): %s", page_title)

    return content, md, page_name, page_title, mtime


def handle_markdown_file_path(path: str) -> str:
    """Given a location on disk, attempt to open it and render the markdown within."""
    content, md, page_name, page_title, mtime = parse_md(path)
    parent_navs = generate_parent_navs(path)
    extra_footer = get_meta_str(md, 'footer') if md.Meta.get('footer') else None
    template_name = get_meta_str(md, 'template') if md.Meta.get('template') else 'base.html'

    # check if this has a HTTP redirect
    redirect_url = get_meta_str(md, 'redirect') if md.Meta.get('redirect') else None
    if redirect_url:
        raise NotImplementedError("redirects in markdown are unsupported!")

    template = jinja_env.get_template(template_name)
    return template.render(title=page_title,
                           config=Config,
                           description=get_meta_str(md, 'description'),
                           image=get_meta_str(md, 'image'),
                           content=content,
                           base_url=Config.BASE_HOST + instance_resource_path_to_request_path(path),
                           navs=parent_navs,
                           mtime=mtime.strftime('%Y-%m-%d %H:%M:%S %Z'),
                           extra_footer=extra_footer)


def generate_parent_navs(path):
    """Create a series of paths/links to navigate up from the given resource path."""
    if path == 'index.md':
        # bail and return the domain name as a terminal case
        return [(Config.DOMAIN_NAME, '/')]
    else:
        if path.endswith('index.md'):
            # index case: one dirname for foo/bar/index.md -> foo/bar, one for foo/bar -> foo
            parent_resource_dir = os.path.dirname(os.path.dirname(path))
        else:
            # usual case: foo/buh.md -> foo
            parent_resource_dir = os.path.dirname(path)

        # generate the request path (i.e. what the link will be) for this path, and
        # also the resource path of this parent (which is always a dir, so always index.md)
        request_path = instance_resource_path_to_request_path(path)
        parent_resource_path = os.path.join(parent_resource_dir, 'index.md')

        logger.debug("resource path: '%s'; request path: '%s'; parent resource path: '%s'", path,
                     request_path, parent_resource_path)

        # for issues regarding parser reuse (see lib.init_md) we reinitialize the parser here
        md = init_md()

        # read the resource
        try:
            with open(path, 'r') as entry_file:
                entry = entry_file.read()
            _ = Markup(md.convert(entry))       # nosec B704
            page_name = (" ".join(md.Meta.get('title')) if md.Meta.get('title')
                         else request_path_to_breadcrumb_display(request_path))
            return generate_parent_navs(parent_resource_path) + [(page_name, request_path)]
        except FileNotFoundError:
            return generate_parent_navs(parent_resource_path) + [(request_path, request_path)]


def request_path_to_breadcrumb_display(path):
    """Given a request path, e.g. "/foo/bar/baz/", turn it into breadcrumby text "baz"."""
    undired = path.rstrip('/')
    leaf = undired[undired.rfind('/'):]
    return leaf.strip('/')



"""Process the error page templates.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os

from incorporealcms import jinja_env
from incorporealcms.config import Config


def generate_error_pages(dest_dir: str) -> None:
    """Process the error pages and place them in the output dir.

    Args:
        dest_dir: the directory to place the error pages in, for a web server to serve
    """
    for template_name in ['400.html', '404.html', '500.html']:
        template = jinja_env.get_template(template_name)
        with open(os.path.join(dest_dir, template_name), 'w') as error_page:
            error_page.write(template.render(config=Config))



"""Test graphviz functionality.

SPDX-FileCopyrightText: © 2021 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os
import tempfile

import pytest

from incorporealcms import init_instance
from incorporealcms.ssg import StaticSiteGenerator

HERE = os.path.dirname(os.path.abspath(__file__))

init_instance(instance_path=os.path.join(HERE, 'instance'),
              extra_config={'MARKDOWN_EXTENSIONS': ['incorporealcms.mdx.pydot', 'incorporealcms.mdx.figures',
                                                    'attr_list']})


def test_graphviz_is_rendered():
    """Initialize the app with the graphviz extension and ensure it does something."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        ssg = StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(src_dir, 'pages'))

        ssg.build_file_in_destination(os.path.join(HERE, 'instance', 'pages'), '', 'test-graphviz.md', tmpdir, True)
        with open(os.path.join(tmpdir, 'test-graphviz.html'), 'r') as graphviz_output:
            data = graphviz_output.read()
            assert 'data:image/png;base64' in data
    os.chdir(HERE)


def test_invalid_graphviz_is_not_rendered():
    """Check that invalid graphviz doesn't blow things up."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        ssg = StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(src_dir, 'broken'))

        with pytest.raises(ValueError):
            ssg.build_file_in_destination(os.path.join(HERE, 'instance', 'broken'), '', 'test-invalid-graphviz.md',
                                          tmpdir, True)
    os.chdir(HERE)


def test_figures_are_rendered():
    """Test that a page with my figure syntax renders as expected."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        ssg = StaticSiteGenerator(src_dir, tmpdir)
        os.chdir(os.path.join(src_dir, 'pages'))

        ssg.build_file_in_destination(os.path.join(HERE, 'instance', 'pages'), '', 'figures.md', tmpdir, True)
        with open(os.path.join(tmpdir, 'figures.html'), 'r') as graphviz_output:
            data = graphviz_output.read()
            assert ('<figure class="right"><img alt="fancy captioned logo" src="bss-square-no-bg.png" />'
                    '<figcaption>this is my cool logo!</figcaption></figure>') in data
            assert ('<figure><img alt="vanilla captioned logo" src="bss-square-no-bg.png" />'
                    '<figcaption>this is my cool logo without an attr!</figcaption>\n</figure>') in data
            assert ('<figure class="left"><img alt="fancy logo" src="bss-square-no-bg.png" />'
                    '<span></span></figure>') in data
            assert '<figure><img alt="just a logo" src="bss-square-no-bg.png" /></figure>' in data
    os.chdir(HERE)



"""Test the feed methods.

SPDX-FileCopyrightText: © 2023 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import os
import tempfile

from incorporealcms import init_instance
from incorporealcms.feed import generate_feed

HERE = os.path.dirname(os.path.abspath(__file__))

init_instance(instance_path=os.path.join(HERE, 'instance'))


def test_atom_type_generated():
    """Test that an ATOM feed can be generated."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generate_feed('atom', src_dir, tmpdir)

        with open(os.path.join(tmpdir, 'feed', 'atom'), 'r') as feed_output:
            data = feed_output.read()
            assert '<?xml version=\'1.0\' encoding=\'UTF-8\'?>\n<feed xmlns="http://www.w3.org/2005/Atom">' in data
            assert '<id>https://example.org/</id>' in data
            assert '<email>admin@example.org</email>' in data
            assert '<name>Test Name</name>' in data

            # forced-no-title.md
            assert '<title>example.org</title>' in data
            assert '<link href="https://example.org/forced-no-title"/>' in data
            assert '<id>tag:example.org,2023-12-01:/forced-no-title</id>' in data
            assert '<content type="html">&lt;p&gt;some words are here&lt;/p&gt;</content>' in data

            # more-metadata.md
            assert '<title>title for the page - example.org</title>' in data
            assert '<link href="https://example.org/more-metadata"/>' in data
            assert '<id>tag:example.org,2025-03-16:/more-metadata</id>' in data
            assert '<content type="html">&lt;p&gt;hello&lt;/p&gt;</content>' in data


def test_rss_type_generated():
    """Test that an RSS feed can be generated."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generate_feed('rss', src_dir, tmpdir)

        with open(os.path.join(tmpdir, 'feed', 'rss'), 'r') as feed_output:
            data = feed_output.read()
            assert '<?xml version=\'1.0\' encoding=\'UTF-8\'?>\n<rss xmlns:atom="http://www.w3.org/2005/Atom"' in data
            assert 'xmlns:content="http://purl.org/rss/1.0/modules/content/" version="2.0">' in data
            assert '<link>https://example.org</link>' in data

            # forced-no-title.md
            assert '<title>example.org</title>' in data
            assert '<link>https://example.org/forced-no-title</link>' in data
            assert '<guid isPermaLink="false">tag:example.org,2023-12-01:/forced-no-title</guid>' in data
            assert '<description>&lt;p&gt;some words are here&lt;/p&gt;</description>' in data
            assert '<author>admin@example.org (Test Name)</author>' in data

            # more-metadata.md
            assert '<title>title for the page - example.org</title>' in data
            assert '<link>https://example.org/more-metadata</link>' in data
            assert '<guid isPermaLink="false">tag:example.org,2025-03-16:/more-metadata</guid>' in data
            assert '<description>&lt;p&gt;hello&lt;/p&gt;</description>' in data
            assert '<author>admin@example.org (Test Name)</author>' in data


def test_multiple_runs_without_error():
    """Test that we can run the RSS and Atom feed generators in any order."""
    with tempfile.TemporaryDirectory() as tmpdir:
        src_dir = os.path.join(HERE, 'instance')
        generate_feed('atom', src_dir, tmpdir)
        generate_feed('rss', src_dir, tmpdir)
        generate_feed('atom', src_dir, tmpdir)
        generate_feed('rss', src_dir, tmpdir)



"""Build an instance as a static site suitable for serving via e.g. Nginx.

SPDX-FileCopyrightText: © 2025 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: GPL-3.0-only
"""
import argparse
import os
import shutil
import stat
import tempfile

from termcolor import cprint

from incorporealcms import __version__, init_instance
from incorporealcms.error_pages import generate_error_pages
from incorporealcms.feed import generate_feed
from incorporealcms.markdown import handle_markdown_file_path


class StaticSiteGenerator(object):
    """Generate static site output based on the instance's content."""

    def __init__(self, instance_dir: str, output_dir: str, extra_config=None):
        """Create the object to run various operations to generate the static site.

        Args:
            instance_dir: the directory from which to read an instance format set of content
            output_dir: the directory to write the generated static site to
        """
        self.instance_dir = os.path.abspath(instance_dir)
        self.output_dir = os.path.abspath(output_dir)

        # initialize configuration with the path to the instance
        init_instance(self.instance_dir, extra_config)

    def build(self):
        """Build the whole static site."""
        # putting the temporary directory next to the desired output so we can safely rename it later
        tmp_output_dir = tempfile.mkdtemp(dir=os.path.dirname(self.output_dir))
        cprint(f"creating temporary directory '{tmp_output_dir}' for writing", 'green')

        # copy core content
        pages_dir = os.path.join(self.instance_dir, 'pages')
        self.build_in_destination(pages_dir, tmp_output_dir)

        # copy the program's static dir
        program_static_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'static')
        static_output_dir = os.path.join(tmp_output_dir, 'static')
        try:
            os.mkdir(static_output_dir)
        except FileExistsError:
            # already exists
            pass
        self.build_in_destination(program_static_dir, static_output_dir, convert_markdown=False)

        # copy the instance's static dir --- should I deprecate this since it could just be stuff in pages/static/?
        custom_static_dir = os.path.join(self.instance_dir, 'custom-static')
        self.build_in_destination(custom_static_dir, static_output_dir, convert_markdown=False)

        # generate the feeds
        cprint("generating feeds", 'green')
        generate_feed('atom', self.instance_dir, tmp_output_dir)
        generate_feed('rss', self.instance_dir, tmp_output_dir)

        # generate the error pages
        cprint("generating error pages", 'green')
        generate_error_pages(tmp_output_dir)

        # move temporary dir to the destination
        old_output_dir = f'{self.output_dir}-old-{os.path.basename(tmp_output_dir)}'
        if os.path.exists(self.output_dir):
            cprint(f"renaming '{self.output_dir}' to '{old_output_dir}'", 'green')
            os.rename(self.output_dir, old_output_dir)
        cprint(f"renaming '{tmp_output_dir}' to '{self.output_dir}'", 'green')
        os.rename(tmp_output_dir, self.output_dir)
        os.chmod(self.output_dir,
                 stat.S_IWUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

        # TODO: unlink old dir above? arg flag?

    def build_in_destination(self, source_dir: str, dest_dir: str, convert_markdown: bool = True) -> None:
        """Walk the source directory and copy and/or convert its contents into the destination.

        Args:
            source_dir: the directory to copy into the destination
            dest_dir: the directory to place copied/converted files into
            convert_markdown: whether or not to convert Markdown files (or simply copy them)
        """
        cprint(f"copying files from '{source_dir}' to '{dest_dir}'", 'green')
        os.chdir(source_dir)
        for base_dir, subdirs, files in os.walk(source_dir):
            # remove the absolute path of the directory from the base_dir
            base_dir = os.path.relpath(base_dir, source_dir)
            # create subdirs seen here for subsequent depth
            for subdir in subdirs:
                self.build_subdir_in_destination(source_dir, base_dir, subdir, dest_dir)

            # process and copy files
            for file_ in files:
                self.build_file_in_destination(source_dir, base_dir, file_, dest_dir, convert_markdown)

    def build_subdir_in_destination(self, source_dir: str, base_dir: str, subdir: str, dest_dir: str) -> None:
        """Create a subdir (which might actually be a symlink) in the output dir.

        Args:
            source_dir: the absolute path of the location in the instance, contains subdir
            base_dir: the relative path of the location in the instance, contains subdir
            subdir: the subdir in the instance to replicate in the output
            dest_dir: the output directory to place the subdir in
        """
        dst = os.path.join(dest_dir, base_dir, subdir)
        if os.path.islink(os.path.join(base_dir, subdir)):
            # keep the link relative to the output directory
            src = self.symlink_to_relative_dest(source_dir, os.path.join(base_dir, subdir))
            print(f"creating directory symlink '{dst}' -> '{src}'")
            os.symlink(src, dst, target_is_directory=True)
        else:
            print(f"creating directory '{dst}'")
            try:
                os.mkdir(dst)
            except FileExistsError:
                # already exists
                pass

    def build_file_in_destination(self, source_dir: str, base_dir: str, file_: str, dest_dir: str,
                                  convert_markdown=False) -> None:
        """Create a file (which might actually be a symlink) in the output dir.

        Args:
            source_dir: the absolute path of the location in the instance, contains subdir
            base_dir: the relative path of the location in the instance, contains subdir
            file_: the file in the instance to replicate in the output
            dest_dir: the output directory to place the subdir in
        """
        dst = os.path.join(dest_dir, base_dir, file_)
        if os.path.islink(os.path.join(base_dir, file_)):
            # keep the link relative to the output directory
            src = self.symlink_to_relative_dest(source_dir, os.path.join(base_dir, file_))
            print(f"creating symlink '{dst}' -> '{src}'")
            os.symlink(src, dst, target_is_directory=False)
        else:
            src = os.path.join(base_dir, file_)
            print(f"copying file '{src}' -> '{dst}'")
            shutil.copy2(src, dst)

            # render markdown as HTML
            if src.endswith('.md') and convert_markdown:
                rendered_file = dst.removesuffix('.md') + '.html'
                try:
                    content = handle_markdown_file_path(src)
                except UnicodeDecodeError:
                    # perhaps this isn't a markdown file at all for some reason; we
                    # copied it above so stick with tha
                    cprint(f"{src} has invalid bytes! skipping", 'yellow')
                else:
                    with open(rendered_file, 'w') as dst_file:
                        dst_file.write(content)

    def symlink_to_relative_dest(self, base_dir: str, source: str) -> str:
        """Given a symlink, make sure it points to something inside the instance and provide its real destination.

        Args:
            base_dir: the full absolute path of the instance's pages dir, which the symlink destination must be in.
            source: the symlink to check
        Returns:
            what the symlink points at
        """
        if not os.path.realpath(source).startswith(base_dir):
            raise ValueError(f"symlink destination {os.path.realpath(source)} is outside the instance!")
        # this symlink points to realpath inside base_dir, so relative to base_dir, the symlink dest is...
        return os.path.relpath(os.path.realpath(source), base_dir)


def build():
    """Build the static site generated against an instance directory."""
    parser = argparse.ArgumentParser(
        description="Build the static site generated against an instance directory.",
    )
    parser.add_argument(
        'instance_dir', help="path to instance directory root (NOTE: the program will go into pages/)"
    )
    parser.add_argument(
        'output_dir', help="path to directory to output to (NOTE: the program must be able to write into its parent!)"
    )
    args = parser.parse_args()

    cprint(f"incorporealcms-build v{__version__} Copyright (C) 2025 Brian S. Stephan <bss@incorporeal.org>", 'green')
    # check output path before doing work
    if not os.path.isdir(args.output_dir):
        # if it doesn't exist, great, we'll just move the temporary dir later;
        # if it exists and is a dir, that's fine, but if it's a file, we should error
        if os.path.exists(args.output_dir):
            raise ValueError(f"specified output path '{args.output_dir}' exists as a file!")

    site_gen = StaticSiteGenerator(args.instance_dir, args.output_dir)
    site_gen.build()
